// sa1200p_with_esp_wifi.ino
// weigu.lu

// Publishes every in milliseconds
const long PUBLISH_TIME = 5000;

// Comment or uncomment the following lines suiting your needs
//#define STATIC        // if static IP needed (no DHCP)
//#define OTA           // if Over The Air update needed (security risk!)

#include "ESPBacker.h"  // ESP helper lib (more on weigu.lu)
#include <PubSubClient.h>
#include <SoftwareSerial.h>

// WiFi and network settings
const char *WIFI_SSID = "***";                 // SSID
const char *WIFI_PASSWORD = "***";   // password
const char *NET_MDNSNAME = "MEISCHSCO2";      // optional (access with myESP.local)
const char *NET_HOSTNAME = "MEISCHSCO2";      // optional

#ifdef STATIC
  IPAddress NET_LOCAL_IP (192,168,1,181);    // 3x optional for static IP
  IPAddress NET_GATEWAY (192,168,1,20);
  IPAddress NET_MASK (255,255,255,0);
#endif // ifdef STATIC*/

// UDP settings
const word UDP_LOG_PORT = 6666;
IPAddress UDP_LOG_PC_IP(192,168,1,50);

// Over The Air update settings
#ifdef OTA
  const char *OTA_NAME = "SmartyReader2";
  const char *OTA_PASS_HASH = "c3be31f8c0878e2a4f007200220ce2ba"; //LetMeOTA!
#endif // #ifdef OTA

// MQTT settings
const char *MQTT_SERVER = "192.168.1.60";
const char *MQTT_CLIENT_ID = "meischs_co2_2345678";
String MQTT_TOPIC = "co2_sc1";
#ifdef MQTTSECURE // http://weigu.lu/tutorials/sensors2bus/06_mqtt/index.html
  const short MQTT_PORT = 8883;  // port for secure communication
  const char *MQTT_USER = "me";
  const char *MQTT_PASS = "myMqttPass12!";
  const char *MQTT_PSK_IDENTITY = "btsiot1";
  const char *MQTT_PSK_KEY = "0123456789abcdef0123"; // hex string without 0x
  WiFiClientMQTTSECURE espClient;
#else
  const short MQTT_PORT = 1883; // clear text = 1883
  WiFiClient espClient;
#endif

ESPBacker B;

PubSubClient mqttClient(espClient);

const byte CO2_RX = 14; // D5
const byte SDA_PIN = 4; // D2
const byte SCL_PIN = 5; // D1
byte event_counter = 0;
int co2;
float sht21_value;
float hum, temp;
const byte buffer_bytes = 8;
volatile byte bit_buffer[buffer_bytes];
volatile byte bit_counter = 0;
String Mqtt_msg = "";

SoftwareSerial SerialCO2(CO2_RX); // RX

void setup() {
  B.set_led_log(true);                 // use builtin LED for debugging
  B.set_serial_log(false);             // true if you need debug info (S0)
  pinMode(SDA_PIN,INPUT);
  pinMode(SCL_PIN,INPUT);
  Serial.begin(115200);
  SerialCO2.begin(9600);
  delay(500);
  Serial.println("Hi");
  attachInterrupt(digitalPinToInterrupt(SCL_PIN), isr_scl, RISING);
  init_wifi();
  mqttClient.setServer(MQTT_SERVER,MQTT_PORT); //open connection to MQTT server
  mqtt_connect();
  #ifdef OTA
    B.init_ota(OTA_NAME, OTA_PASS_HASH);
  #endif // ifdef OTA
  noInterrupts();
  clear_buffer(buffer_bytes);
  bit_counter = 0;
  interrupts(); // allow interrupts
}

void loop() {
  #ifdef OTA
    ArduinoOTA.handle();
  #endif // ifdef OTA
  // Publish every PUBLISH_TIME
  if (B.non_blocking_delay(PUBLISH_TIME)) {
    Mqtt_msg = "{\"CO2 [ppm]\":" + String(co2) + ",\"Temperature  [°C]\":" + \
               String(temp) +  ",\"Humidity [%]\":" + String(hum) + "}";
    mqttClient.publish(MQTT_TOPIC.c_str(), Mqtt_msg.c_str());
  }
  if (SerialCO2.available()) {
    co2 = get_CO2();
    if (co2 == -1) {
      Serial.println("Serial data (co2) corrupt");
    }
    else {
      Serial.println("CO2 = " + String(co2) + " ppm");
    }
  }
  sht21_value = get_sht21_data();
  if ((event_counter == 2) && (sht21_value > 0)) {
    hum = sht21_value;
    Serial.println("Humidity is: " + String(hum) + "%");
  }
  else if ((event_counter == 3) && (sht21_value > 0)) {
    temp = sht21_value;
    Serial.println("Temperature is: " + String(temp) + "°C");
    event_counter = 0;
  }
  else if ((sht21_value == -1)) {
    Serial.println("SHT21 CRC error");
    temp = -1;
    hum = -1;
  }
  if (WiFi.status() != WL_CONNECTED) {  // if WiFi disconnected, reconnect
    init_wifi();
  }
  if (!mqttClient.connected()) { // reconnect mqtt client, if needed
    mqtt_connect();
  }
  mqttClient.loop();                    // make the MQTT live
  delay(100);
  yield();
}

ICACHE_RAM_ATTR void isr_scl() {
  //delayMicroseconds(15);
  if (digitalRead(SDA_PIN)) {
    bitSet(bit_buffer[bit_counter/8], 7-bit_counter%8);
  }
  else {
    bitClear(bit_buffer[bit_counter/8], 7-bit_counter%8);
  }
  bit_counter++;
}

void init_wifi() {
  #ifdef STATIC
    B.init_wifi_sta(WIFI_SSID, WIFI_PASSWORD, NET_HOSTNAME, NET_LOCAL_IP,
                  NET_GATEWAY, NET_MASK);
  #else
    B.init_wifi_sta(WIFI_SSID, WIFI_PASSWORD, NET_MDNSNAME, NET_HOSTNAME);
  #endif // ifdef STATIC
}

// connect to MQTT server
void mqtt_connect() {
  while (!mqttClient.connected()) { // Loop until we're reconnected
    B.log("Attempting MQTT connection...");
    #ifdef MQTTSECURE
      if (mqttClient.connect(MQTT_CLIENT_ID, MQTT_USER, MQTT_PASS)) {
    #else
      if (mqttClient.connect(MQTT_CLIENT_ID)) { // Attempt to connect
    #endif // ifdef UNMQTTSECURE
      B.log_ln("MQTT connected");
      // Once connected, publish an announcement...
      //mqttClient.publish(MQTT_TOPIC, "{\"dt\":\"connected\"}");
      // don't because OpenHAB does not like this ...
    }
    else {
      B.log("MQTT connection failed, rc=");
      B.log(String(mqttClient.state()));
      B.log_ln(" try again in 5 seconds");
      delay(5000);  // Wait 5 seconds before retrying
    }
  }
}


//clear buffer
void clear_buffer(byte lbuffer_bytes) {
  for (byte i=0; i<lbuffer_bytes; i++) {
    bit_buffer[i]=0;
  }
}

// get the CO2 data
int get_CO2() {
  int co2_lvalue;
  byte co2_byte_counter = 0, co2_data[10], co2_checksum;
  while (SerialCO2.available()) {
    co2_data[co2_byte_counter] = SerialCO2.read();
    co2_byte_counter++;
  }
  // are the three first byte ok?
  if ((co2_data[0] != 0x16) || (co2_data[1] != 0x05) || (co2_data[2] != 0x01)) {
    B.log_ln("data corrupt");
    return -1;
  }
  // calculate the checksum and test if ok
  co2_checksum = 0;
  for (byte i=0; i<co2_byte_counter-1; i++) {
    co2_checksum += co2_data[i];
  }
  co2_checksum = 256-co2_checksum;
  if (co2_data[co2_byte_counter-1]!=co2_checksum) {
    B.log_ln("Checksum not ok");
    return -1;
  }
  co2_lvalue = co2_data[3]*256+co2_data[4];
  for (byte i=0; i<co2_byte_counter; i++) {
    B.log(String(co2_data[i],HEX));
    B.log(" ");
  }
  B.log(" checksum: ");
  B.log_ln(String(co2_checksum,HEX));
  for (byte i=0; i<sizeof(co2_data); i++) { //clear buffer
    co2_data[i]=0;
  }
  co2_byte_counter = 0;
  return co2_lvalue;
}

// we need global vars bit_counter, bitbuffer, envent_counter
float get_sht21_data() {
  byte msb, lsb, crc;
  unsigned int data_word = 0;
  float lhum, ltemp;
  if (bit_counter >= 28) {
    delay(2); //wait 2ms for the other bits
    noInterrupts();
    if (bit_buffer[0]==0x80) { // master
      event_counter = 1;
    }
    else if (event_counter == 1) { // hum data from slave
      msb = bit_buffer[0];
      lsb = bit_buffer[1] << 1;    // eliminate ACK bits
      crc = bit_buffer[2] << 2;
      lsb = lsb + (bit_buffer[2] >> 7);
      crc = crc + (bit_buffer[3] >> 6);
      if (check_crc(msb, lsb, crc) == false) {
        event_counter = 0;
        interrupts();
        return -1;
      }
      lsb &= 0xFC;                 // clear last 2 bits (data sheet)
      data_word = msb*256+lsb;
      lhum = -6.0+(125.0*data_word/65536.0);
      B.log("msb: " + String(msb,HEX) + " lsb: " + String(lsb,HEX));
      B.log_ln(" crc: " + String(crc,HEX) + " data word: " + String(data_word));
      event_counter = 2;
    }
    else if (event_counter == 2) { // temp data from slave
            msb = bit_buffer[0];
      lsb = bit_buffer[1] << 1;    // eliminate ACK bits
      crc = bit_buffer[2] << 2;
      lsb = lsb + (bit_buffer[2] >> 7);
      crc = crc + (bit_buffer[3] >> 6);
      if (check_crc(msb, lsb, crc) == false) {
        event_counter = 0;
        interrupts();
        return -1;
      }
      lsb &= 0xFC;                 // clear last 2 bits (data sheet)
      data_word = msb*256+lsb;
      ltemp = -46.85+(175.72*data_word/65536.0);
      ltemp = ltemp -1.65;          // correction for housing error?
      B.log("msb: " + String(msb,HEX) + " lsb: " + String(lsb,HEX));
      B.log_ln(" crc: " + String(crc,HEX) + " data word: " + String(data_word));
      event_counter = 3;
    }
    B.log(String(event_counter) + ":  " + String(bit_counter) + " bit:  ");
    for (byte i=0; i<buffer_bytes; i++) {
      B.log(String(bit_buffer[i],HEX));
      B.log("\t");
    }
    delay(10);
    B.log_ln();
    clear_buffer(buffer_bytes);
    bit_counter = 0;
    if (event_counter == 2) {
      interrupts();
      return lhum;
    }
    else if (event_counter == 3) {
      interrupts();
      return ltemp;
    }
    interrupts();
  }
}

// from Sensiron application note and code
bool check_crc(byte lmsb, byte llsb, byte lcrc) {
  byte bitmask;
  byte calc_crc = 0x00; // initial value as per Table 17
  calc_crc ^= (lmsb); // do msb
  for (bitmask = 8; bitmask > 0; --bitmask)  {
    if (calc_crc & 0x80) {
      calc_crc = (calc_crc << 1) ^ 0x131;
    }
    else {
      calc_crc = (calc_crc << 1);
    }
  }
  calc_crc ^= (llsb); // do lsb
  for (bitmask = 8; bitmask > 0; --bitmask)  {
    if (calc_crc & 0x80) {
      calc_crc = (calc_crc << 1) ^ 0x131; // polynomial from Table 17
    }
    else {
      calc_crc = (calc_crc << 1);
    }
  }
  B.log_ln("------------------");
  B.log("msb: " + String(lmsb,HEX) + " lsb: " + String(llsb,HEX) + " crc: ");
  B.log_ln(String(lcrc,HEX) + "  calculated checksum = " + String(calc_crc,HEX));
  B.log_ln("------------------");
  if (calc_crc != lcrc) {
    return false;     // checksum error
  }
  else {
    return true;      // no error
  }
}
